/*
 * Copyright (C) 2012 - Michael Baudin
 * Copyright (C) Gaüzère Sabine
 *
 * This file must be used under the terms of the CeCILL.
 * This source file is licensed as described in the file COPYING, which
 * you should have received as part of this distribution.  The terms
 * are also available at
 * http://www.cecill.info/licences/Licence_CeCILL_V2-en.txt
 *
 */

/*
This source code was taken in the project "freemat" (BSD license)
*/
#include <math.h>

#include "cdflib.h"
#include "cdflib_private.h"


/*
lambda=MEAN lambda OF THE POISSON DISTRIBUTION
OUTPUT: cdflib_poissrnd=SAMPLE FROM THE POISSON-(lambda)-DISTRIBUTION
lambdaprev=PREVIOUS lambda, lambdaold=lambda AT LAST EXECUTION OF STEP P OR B.
TABLES: COEFFICIENTS A0-A7 FOR STEP F. FACTORIALS FACT
COEFFICIENTS A(K) - FOR PX = FK*V*V*SUM(A(K)*V**K)-DEL
SEPARATION OF CASES A AND B
*/
int cdflib_poissrnd(double lambda, double *x)
{
	static double a0 = -0.5;
	static double a1 = 0.3333333;
	static double a2 = -0.2500068;
	static double a3 = 0.2000118;
	static double a4 = -0.1661269;
	static double a5 = 0.1421878;
	static double a6 = -0.1384794;
	static double a7 = 0.125006;
	static double lambdaold = -1.0E37;
	static double lambdaprev = -1.0E37;
	static double fact[10] = {
		1.0,1.0,2.0,6.0,24.0,120.0,720.0,5040.0,40320.0,362880.0
	};
	static int j,k,kflag,l,m,ll;     
	// JJV I added a variable 'll' here - it is the 'l' for CASE A
	static double b1,b2,c,c0,c1,c2,c3,d,del,difmuk,e,fk,fx,fy,g,omega,p,p0,px,py,q,s,
		t,u,v,y,xx,pp[35];
	double R;
	double z;
	// This is 1/sqrt(2*%pi)
	double c1_SQRT_2PI = 0.3989422804014326779399;
	// This is 1/24
	double c1_24 = 0.0416666666666666666666666666666666666666666666666666;
	// This is 1/7
	double c1_7=0.1428571428571428571428571428571428571428571428571428;
	double c1_12=0.0833333333333333333333333333333333333333333333333333;

	if(lambda == lambdaprev) {
		goto S10;
	}
	if(lambda < 10.0) {
		goto S120;
	}
	/*
	CASE  A. (RECALCULATION OF S,D,L IF lambda HAS CHANGED)
	*/
	//   JJV This is the case where I changed 'l' to 'll'
	//   JJV Here 'll' is set once and used in a comparison once
	lambdaprev = lambda;
	s = sqrt(lambda);
	d = 6.0* lambda * lambda;
	/*
	THE POISSON PROBABILITIES PK EXCEED THE DISCRETE NORMAL
	PROBABILITIES FK WHENEVER K >= M(lambda). L=IFIX(lambda-1.1484)
	IS AN UPPER BOUND TO M(lambda) FOR ALL lambda >= 10 .
	*/
	ll = (int) (lambda-1.1484);
S10:
	/*
	STEP N. NORMAL SAMPLE - cdflib_snorm(IR) FOR STANDARD NORMAL DEVIATE
	*/
	z=cdflib_snorm();
	g = lambda+s*z;
	if(g < 0.0) {
		goto S20;
	}
	R = floor(g);
	/*
	STEP I. IMMEDIATE ACCEPTANCE IF R IS LARGE ENOUGH
	*/
	if(R >= ll) {
		*x=R;
		return CDFLIB_OK;
	}
	/*
	STEP S. SQUEEZE ACCEPTANCE - SUNIF(IR) FOR (0,1)-SAMPLE U
	*/
	fk = R;
	difmuk = lambda-fk;
	u = cdflib_randgenerate();
	if(d*u >= difmuk*difmuk*difmuk) {
		*x=R;
		return CDFLIB_OK;
	}
S20:
	/*
	STEP P. PREPARATIONS FOR STEPS Q AND H.
	(RECALCULATIONS OF PARAMETERS IF NECESSARY)
	.3989423=(2*PI)**(-.5)  .416667E-1=1./24.  .1428571=1./7.
	THE QUANTITIES B1, B2, C3, C2, C1, C0 ARE FOR THE HERMITE
	APPROXIMATIONS TO THE DISCRETE NORMAL PROBABILITIES FK.
	C=.1069/lambda GUARANTEES MAJORIZATION BY THE 'HAT'-FUNCTION.
	*/
	if(lambda == lambdaold) {
		goto S30;
	}
	lambdaold = lambda;
	omega = c1_SQRT_2PI/s;
	b1 = c1_24/ lambda;
	b2 = 0.3*b1*b1;
	c3 = c1_7*b1*b2;
	c2 = b2-15.0*c3;
	c1 = b1-6.0*b2+45.0*c3;
	c0 = 1.0-b1+3.0*b2-15.0*c3;
	c = 0.1069/ lambda;
S30:
	if(g < 0.0) {
		goto S50;
	}
	/*
	'SUBROUTINE' F IS CALLED (KFLAG=0 FOR CORRECT RETURN)
	*/
	kflag = 0;
	goto S70;
S40:
	/*
	STEP Q. QUOTIENT ACCEPTANCE (RARE CASE)
	*/
	if(fy-u*fy <= py*exp(px-fx)) {
		*x=R;
		return CDFLIB_OK;
	}
S50:
	/*
	STEP E. EXPONENTIAL SAMPLE - cdflib_sexpo() FOR STANDARD EXPONENTIAL
	DEVIATE E AND SAMPLE T FROM THE LAPLACE 'HAT'
	(IF T <= -.6744 THEN PK < FK FOR ALL lambda >= 10.)
	*/
	e = cdflib_sexpo();
	u = cdflib_randgenerate();
	u += (u-1.0);
	t = 1.8+cdflib_dsign(e,u);
	if(t <= -0.6744) {
		goto S50;
	}
	R = floor(lambda+s*t);
	fk = R;
	difmuk = lambda-fk;
	/*
	'SUBROUTINE' F IS CALLED (KFLAG=1 FOR CORRECT RETURN)
	*/
	kflag = 1;
	goto S70;
S60:
	/*
	STEP H. HAT ACCEPTANCE (E IS REPEATED ON REJECTION)
	*/
	if(c*Abs(u) > py*exp(px+e)-fy*exp(fx+e)) {
		goto S50;
	}
	*x=R;
	return CDFLIB_OK;
S70:
	/*
	STEP F. 'SUBROUTINE' F. CALCULATION OF PX,PY,FX,FY.
	CASE R .LT. 10 USES FACTORIALS FROM TABLE FACT
	*/
	if(R >= 10) {
		goto S80;
	}
	px = - lambda;
	py = pow(lambda,R)/ fact[(int)R];
	goto S110;
S80:
	/*
	CASE R .GE. 10 USES POLYNOMIAL APPROXIMATION
	A0-A7 FOR ACCURACY WHEN ADVISABLE
	.8333333E-1=1./12.  .3989423=(2*PI)**(-.5)
	*/
	del = c1_12/fk;
	del -= (4.8*del*del*del);
	v = difmuk/fk;
	if(Abs(v) <= 0.25) {
		goto S90;
	}
	px = fk*log(1.0+v)-difmuk-del;
	goto S100;
S90:
	px = fk*v*v*(((((((a7*v+a6)*v+a5)*v+a4)*v+a3)*v+a2)*v+a1)*v+a0)-del;
S100:
	py = c1_SQRT_2PI/sqrt(fk);
S110:
	y = (0.5-difmuk)/s;
	xx = y*y;
	fx = -0.5*xx;
	fy = omega*(((c3*xx+c2)*xx+c1)*xx+c0);
	if(kflag <= 0) {
		goto S40;
	}
	goto S60;
S120:
	/*
	CASE  B. (START NEW TABLE AND CALCULATE P0 IF NECESSARY)
	*/
	lambdaprev = -1.0E37;
	m = max(1L,(int) (lambda));
	l = 0;
	p = exp(- lambda);
	q = p0 = p;
S130:
	/*
	STEP U. UNIFORM SAMPLE FOR INVERSION METHOD
	*/
	u = cdflib_randgenerate();
	R = 0;
	if(u <= p0) {
		*x=R;
		return CDFLIB_OK;
	}
	/*
	STEP T. TABLE COMPARISON UNTIL THE END PP(L) OF THE
	PP-TABLE OF CUMULATIVE POISSON PROBABILITIES
	(0.458=PP(9) FOR lambda=10)
	*/
	if(l == 0) {
		goto S150;
	}
	j = 1;
	if(u > 0.458) {
		j = min(l,m);
	}
	for(k=j; k<=l; k++) {
		if(u <= *(pp+k-1)) {
			goto S180;
		}
	}
	if(l == 35) {
		goto S130;
	}
S150:
	/*
	STEP C. CREATION OF NEW POISSON PROBABILITIES P
	AND THEIR CUMULATIVES Q=PP(K)
	*/
	l += 1;
	for(k=l; k<=35; k++) {
		p = p* lambda/(double)k;
		q += p;
		*(pp+k-1) = q;
		if(u <= q) {
			goto S170;
		}
	}
	l = 35;
	goto S130;
S170:
	l = k;
S180:
	R = k;
	*x=R;
	return CDFLIB_OK;
}
